function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
import { serializeCase } from "../nodes/discriminate.js";
import { checkClass } from "../nodes/rules/class.js";
import { checkDivisor } from "../nodes/rules/divisor.js";
import { checkBound } from "../nodes/rules/range.js";
import { checkRegex } from "../nodes/rules/regex.js";
import { precedenceMap } from "../nodes/rules/rules.js";
import { domainOf, hasDomain } from "../utils/domains.js";
import { hasKey, objectKeysOf } from "../utils/generics.js";
import { getPath, Path } from "../utils/paths.js";
import { deserializePrimitive } from "../utils/serialize.js";
import { defaultProblemWriters, Problem, Problems } from "./problems.js";
var _seen = /*#__PURE__*/ new WeakMap();
export class TraversalState {
    getConfigForProblemCode(code) {
        if (!this.configs.length) {
            return defaultProblemWriters[code];
        }
        for(let i = this.configs.length - 1; i >= 0; i--){
            if (this.configs[i][code] || this.configs[i]["defaults"]) {
                return {
                    ...defaultProblemWriters[code],
                    ...this.configs[i]["defaults"],
                    ...this.configs[i][code]
                };
            }
        }
        return defaultProblemWriters[code];
    }
    traverseKey(key, node) {
        const lastData = this.data;
        this.data = this.data[key];
        this.path.push(key);
        const isValid = traverse(node, this);
        this.path.pop();
        lastData[key] = this.data;
        this.data = lastData;
        return isValid;
    }
    traverseResolution(name) {
        const resolution = this.type.meta.scope.resolve(name);
        const id = resolution.meta.id;
        // this assignment helps with narrowing
        const data = this.data;
        const isObject = hasDomain(data, "object");
        if (isObject) {
            if (hasKey(_classPrivateFieldGet(this, _seen), id)) {
                if (_classPrivateFieldGet(this, _seen)[id].includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return true;
                }
                _classPrivateFieldGet(this, _seen)[id].push(data);
            } else {
                _classPrivateFieldGet(this, _seen)[id] = [
                    data
                ];
            }
        }
        const lastResolution = this.type;
        this.type = resolution;
        const isValid = traverse(resolution.flat, this);
        this.type = lastResolution;
        if (isObject) {
            _classPrivateFieldGet(this, _seen)[id].pop();
        }
        return isValid;
    }
    traverseBranches(branches) {
        const lastFailFast = this.failFast;
        this.failFast = true;
        const lastProblems = this.problems;
        const branchProblems = new Problems(this);
        this.problems = branchProblems;
        const lastPath = this.path;
        let hasValidBranch = false;
        for (const branch of branches){
            this.path = new Path();
            if (checkEntries(branch, this)) {
                hasValidBranch = true;
                break;
            }
        }
        this.path = lastPath;
        this.problems = lastProblems;
        this.failFast = lastFailFast;
        return hasValidBranch || this.problems.create("branches", branchProblems);
    }
    constructor(data, type){
        _classPrivateFieldInit(this, _seen, {
            writable: true,
            value: void 0
        });
        this.data = data;
        this.type = type;
        this.path = new Path();
        this.problems = new Problems(this);
        this.failFast = false;
        _classPrivateFieldSet(this, _seen, {});
        this.configs = type.meta.scope.config ? [
            type.meta.scope.config
        ] : [];
    }
}
export const traverse = (node, state)=>typeof node === "string" ? domainOf(state.data) === node || state.problems.create("domain", node) : checkEntries(node, state);
export const checkEntries = (entries, state)=>{
    let isValid = true;
    for(let i = 0; i < entries.length; i++){
        const [k, v] = entries[i];
        const entryAllowsData = entryCheckers[k](v, state);
        isValid && (isValid = entryAllowsData);
        if (!isValid) {
            if (state.failFast) {
                return false;
            }
            if (i < entries.length - 1 && precedenceMap[k] < precedenceMap[entries[i + 1][0]]) {
                // if we've encountered a problem, there is at least one entry
                // remaining, and the next entry is of a higher precedence level
                // than the current entry, return immediately
                return false;
            }
        }
    }
    return isValid;
};
export const checkRequiredProp = (prop, state)=>{
    if (prop[0] in state.data) {
        return state.traverseKey(prop[0], prop[1]);
    }
    return state.problems.create("missing", undefined, {
        path: state.path.concat(prop[0]),
        data: undefined
    });
};
const entryCheckers = {
    regex: checkRegex,
    divisor: checkDivisor,
    domains: (domains, state)=>{
        const entries = domains[domainOf(state.data)];
        return entries ? checkEntries(entries, state) : state.problems.create("domainBranches", objectKeysOf(domains));
    },
    domain: (domain, state)=>domainOf(state.data) === domain || state.problems.create("domain", domain),
    bound: checkBound,
    optionalProp: (prop, state)=>{
        if (prop[0] in state.data) {
            return state.traverseKey(prop[0], prop[1]);
        }
        return true;
    },
    // these checks work the same way, the keys are only distinct so that
    // prerequisite props can have a higher precedence
    requiredProp: checkRequiredProp,
    prerequisiteProp: checkRequiredProp,
    indexProp: (node, state)=>{
        if (!Array.isArray(state.data)) {
            return state.problems.create("class", "Array");
        }
        let isValid = true;
        for(let i = 0; i < state.data.length; i++){
            isValid && (isValid = state.traverseKey(`${i}`, node));
            if (!isValid && state.failFast) {
                return false;
            }
        }
        return isValid;
    },
    branches: (branches, state)=>state.traverseBranches(branches),
    switch: (rule, state)=>{
        const dataAtPath = getPath(state.data, rule.path);
        const caseKey = serializeCase(rule.kind, dataAtPath);
        if (hasKey(rule.cases, caseKey)) {
            return checkEntries(rule.cases[caseKey], state);
        }
        const caseKeys = objectKeysOf(rule.cases);
        const missingCasePath = state.path.concat(rule.path);
        return rule.kind === "value" ? state.problems.create("valueBranches", caseKeys.map((k)=>deserializePrimitive(k)), {
            path: missingCasePath,
            data: dataAtPath
        }) : state.problems.create("domainBranches", caseKeys, {
            path: missingCasePath,
            data: dataAtPath
        });
    },
    alias: (name, state)=>state.traverseResolution(name),
    class: checkClass,
    narrow: (narrow, state)=>narrow(state.data, state.problems),
    config: ({ config , node  }, state)=>{
        state.configs.push(config);
        const result = traverse(node, state);
        state.configs.pop();
        return result;
    },
    value: (value, state)=>state.data === value || state.problems.create("value", value),
    morph: (morph, state)=>{
        const out = morph(state.data);
        if (out instanceof Problem) {
            return state.problems.add(out);
        }
        state.data = out;
        return true;
    }
};
