import type { FlatBound } from "../nodes/rules/range.js";
import { DataWrapper } from "../utils/data.js";
import type { Domain } from "../utils/domains.js";
import type { arraySubclassToReadonly, constructor, evaluate, instanceOf, RegexLiteral } from "../utils/generics.js";
import type { DefaultObjectKind } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import type { ConstrainedRuleTraversalData, TraversalState } from "./traverse.js";
export declare class ArkTypeError extends TypeError {
    cause: Problems;
    constructor(problems: Problems);
}
export declare class Problem<code extends ProblemCode = any> {
    code: code;
    path: Path;
    private data;
    private source;
    private writers;
    parts?: Problem[];
    constructor(code: code, path: Path, data: ProblemData<code>, source: ProblemSource<code>, writers: ProblemWriters<code>);
    toString(): string;
    get message(): string;
    get reason(): string;
    get mustBe(): string;
}
export type AddProblemOptions<code extends ProblemCode = ProblemCode> = {
    data?: ProblemData<code>;
    path?: Path;
};
declare class ProblemArray extends Array<Problem> {
    #private;
    byPath: Record<string, Problem>;
    constructor(state: TraversalState);
    create<code extends ProblemCode>(code: code, source: ProblemSource<code>, opts?: AddProblemOptions<code>): false;
    add(problem: Problem): false;
    get summary(): string;
    toString(): string;
    throw(): never;
}
export declare const Problems: new (state: TraversalState) => arraySubclassToReadonly<ProblemArray>;
export type Problems = instanceOf<typeof Problems>;
export declare const describeDomains: (domains: Domain[]) => string;
export declare const describeBranches: (descriptions: string[]) => string;
type ProblemSources = {
    divisor: number;
    class: DefaultObjectKind | constructor;
    domain: Domain;
    domainBranches: Domain[];
    missing: undefined;
    bound: FlatBound;
    regex: RegexLiteral;
    value: unknown;
    valueBranches: unknown[];
    multi: Problem[];
    branches: readonly Problem[];
};
export type ProblemCode = evaluate<keyof ProblemSources>;
export type ProblemSource<code extends ProblemCode = ProblemCode> = ProblemSources[code];
type ProblemDataByCode = {
    [code in ProblemCode]: code extends keyof ConstrainedRuleTraversalData ? ConstrainedRuleTraversalData[code] : unknown;
};
export type ProblemData<code extends ProblemCode = ProblemCode> = ProblemDataByCode[code];
export type MustBeWriter<code extends ProblemCode> = (source: ProblemSources[code]) => string;
export type ReasonWriter<code extends ProblemCode = ProblemCode> = (mustBe: string, data: DataWrapper<code extends keyof ConstrainedRuleTraversalData ? ConstrainedRuleTraversalData[code] : unknown>) => string;
export type ContextWriter = (reason: string, path: Path) => string;
export declare const defaultProblemWriters: DefaultProblemsWriters;
export type ProblemOptions<code extends ProblemCode = ProblemCode> = {
    mustBe?: MustBeWriter<code>;
    writeReason?: ReasonWriter<code>;
    addContext?: ContextWriter;
};
export type ProblemsConfig = evaluate<{
    defaults?: ProblemOptions<ProblemCode>;
} & {
    [code in ProblemCode]?: ProblemOptions<code>;
}>;
export type DefaultProblemsWriters = {
    [code in ProblemCode]: ProblemWriters<code>;
};
export type ProblemWriters<code extends ProblemCode> = Required<ProblemOptions<code>>;
export {};
